@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2018 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 # All rights reserved.
 #
 # License: GPL (version 3 or any later version).
@@ -7,7 +7,6 @@
 # --- END COPYRIGHT BLOCK ---
 
 import re
-import logging
 import os
 import json
 import ldap
@@ -16,8 +15,8 @@ from shutil import copyfile
 from getpass import getpass
 from lib389._constants import ReplicaRole, DSRC_HOME
 from lib389.cli_base.dsrc import dsrc_to_repl_monitor
-from lib389.utils import is_a_dn, copy_with_permissions
-from lib389.replica import Replicas, ReplicationMonitor, BootstrapReplicationManager, Changelog5, ChangelogLDIF
+from lib389.utils import is_a_dn, copy_with_permissions, ds_supports_new_changelog
+from lib389.replica import Replicas, ReplicationMonitor, BootstrapReplicationManager, Changelog5, ChangelogLDIF, Changelog
 from lib389.tasks import CleanAllRUVTask, AbortCleanAllRUVTask
 from lib389._mapped_object import DSLdapObjects
 
@@ -37,7 +36,6 @@ arg_to_attr = {
         'cl_dir': 'nsslapd-changelogdir',
         'max_entries': 'nsslapd-changelogmaxentries',
         'max_age': 'nsslapd-changelogmaxage',
-        'compact_interval': 'nsslapd-changelogcompactdb-interval',
         'trim_interval': 'nsslapd-changelogtrim-interval',
         'encrypt_algo': 'nsslapd-encryptionalgorithm',
         'encrypt_key': 'nssymmetrickey',
@@ -48,6 +46,10 @@ arg_to_attr = {
         'bind_dn': 'nsds5replicabinddn',
         'bind_passwd': 'nsds5replicacredentials',
         'bind_method': 'nsds5replicabindmethod',
+        'bootstrap_conn_protocol': 'nsds5replicabootstraptransportinfo',
+        'bootstrap_bind_dn': 'nsds5replicabootstrapbinddn',
+        'bootstrap_bind_passwd': 'nsds5replicabootstrapcredentials',
+        'bootstrap_bind_method': 'nsds5replicabootstrapbindmethod',
         'frac_list': 'nsds5replicatedattributelist',
         'frac_list_total': 'nsds5replicatedattributelisttotal',
         'strip_list': 'nsds5replicastripattrs',
@@ -132,7 +134,7 @@ def enable_replication(inst, basedn, log, args):
     role = args.role.lower()
     rid = args.replica_id
 
-    if role == "master":
+    if role == "supplier":
         repl_type = '3'
         repl_flag = '1'
     elif role == "hub":
@@ -143,7 +145,7 @@ def enable_replication(inst, basedn, log, args):
         repl_flag = '0'
     else:
         # error - unknown type
-        raise ValueError("Unknown replication role ({}), you must use \"master\", \"hub\", or \"consumer\"".format(role))
+        raise ValueError("Unknown replication role ({}), you must use \"supplier\", \"hub\", or \"consumer\"".format(role))
 
     # Start the propeties and update them as needed
     repl_properties = {
@@ -154,12 +156,12 @@ def enable_replication(inst, basedn, log, args):
         'nsDS5ReplicaId': '65535'
         }
 
-    # Validate master settings
-    if role == "master":
+    # Validate supplier settings
+    if role == "supplier":
         # Do we have a rid?
         if not args.replica_id or args.replica_id is None:
-            # Error, master needs a rid TODO
-            raise ValueError('You must specify the replica ID (--replica-id) when enabling a \"master\" replica')
+            # Error, supplier needs a rid TODO
+            raise ValueError('You must specify the replica ID (--replica-id) when enabling a \"supplier\" replica')
 
         # is it a number?
         try:
@@ -181,14 +183,15 @@ def enable_replication(inst, basedn, log, args):
         repl_properties['nsDS5ReplicaBindDN'] = args.bind_dn
 
     # First create the changelog
-    cl = Changelog5(inst)
-    try:
-        cl.create(properties={
-            'cn': 'changelog5',
-            'nsslapd-changelogdir': inst.get_changelog_dir()
-        })
-    except ldap.ALREADY_EXISTS:
-        pass
+    if not ds_supports_new_changelog():
+        cl = Changelog5(inst)
+        try:
+            cl.create(properties={
+                'cn': 'changelog5',
+                'nsslapd-changelogdir': inst.get_changelog_dir()
+            })
+        except ldap.ALREADY_EXISTS:
+            pass
 
     # Finally enable replication
     replicas = Replicas(inst)
@@ -241,14 +244,14 @@ def promote_replica(inst, basedn, log, args):
     replica = replicas.get(args.suffix)
     role = args.newrole.lower()
 
-    if role == 'master':
-        newrole = ReplicaRole.MASTER
+    if role == 'supplier':
+        newrole = ReplicaRole.SUPPLIER
         if args.replica_id is None:
-            raise ValueError("You need to provide a replica ID (--replica-id) to promote replica to a master")
+            raise ValueError("You need to provide a replica ID (--replica-id) to promote replica to a supplier")
     elif role == 'hub':
         newrole = ReplicaRole.HUB
     else:
-        raise ValueError("Invalid role ({}), you must use either \"master\" or \"hub\"".format(role))
+        raise ValueError("Invalid role ({}), you must use either \"supplier\" or \"hub\"".format(role))
 
     replica.promote(newrole, binddn=args.bind_dn, binddn_group=args.bind_group_dn, rid=args.replica_id)
     log.info("Successfully promoted replica to \"{}\"".format(role))
@@ -365,9 +368,16 @@ def set_repl_config(inst, basedn, log, args):
 
 def get_repl_monitor_info(inst, basedn, log, args):
     connection_data = dsrc_to_repl_monitor(DSRC_HOME, log)
+    credentials_cache = {}
 
     # Additional details for the connections to the topology
     def get_credentials(host, port):
+        # credentials_cache is nonlocal to refer to the instance
+        # from enclosing function (get_repl_monitor_info)`
+        nonlocal credentials_cache
+        key = f'{host}:{port}'
+        if key in credentials_cache:
+            return credentials_cache[key]
         found = False
         if args.connections:
             connections = args.connections
@@ -402,8 +412,10 @@ def get_repl_monitor_info(inst, basedn, log, args):
             binddn = input(f'\nEnter a bind DN for {host}:{port}: ').rstrip()
             bindpw = getpass(f"Enter a password for {binddn} on {host}:{port}: ").rstrip()
 
-        return {"binddn": binddn,
-                "bindpw": bindpw}
+        credentials = {"binddn": binddn,
+                       "bindpw": bindpw}
+        credentials_cache[key] = credentials
+        return credentials
 
     repl_monitor = ReplicationMonitor(inst)
     report_dict = repl_monitor.generate_report(get_credentials, args.json)
@@ -434,15 +446,16 @@ def get_repl_monitor_info(inst, basedn, log, args):
             log.info(supplier_header)
 
         # Draw a line with the same length as the header
-        status = ""
         if not args.json:
             log.info("-".join(["" for _ in range(0, len(supplier_header)+1)]))
-        if report_data[0]["replica_status"].startswith("Unavailable"):
-            status = report_data[0]["replica_status"]
-            if not args.json:
-                log.info(f"Replica Status: {status}\n")
-        else:
-            for replica in report_data:
+
+        for replica in report_data:
+            if replica["replica_status"].startswith("Unreachable") or \
+                replica["replica_status"].startswith("Unavailable"):
+                status = replica["replica_status"]
+                if not args.json:
+                    log.info(f"Replica Status: {status}\n")
+            else:
                 replica_root = replica["replica_root"]
                 replica_id = replica["replica_id"]
                 replica_status = replica["replica_status"]
@@ -463,7 +476,7 @@ def get_repl_monitor_info(inst, basedn, log, args):
     if args.json:
         log.info(json.dumps({"type": "list", "items": report_items}, indent=4))
 
-
+# This subcommand is available when 'not ds_supports_new_changelog'
 def create_cl(inst, basedn, log, args):
     cl = Changelog5(inst)
     try:
@@ -476,6 +489,7 @@ def create_cl(inst, basedn, log, args):
     log.info("Successfully created replication changelog")
 
 
+# This subcommand is available when 'not ds_supports_new_changelog'
 def delete_cl(inst, basedn, log, args):
     cl = Changelog5(inst)
     try:
@@ -485,6 +499,7 @@ def delete_cl(inst, basedn, log, args):
     log.info("Successfully deleted replication changelog")
 
 
+# This subcommand is available when 'not ds_supports_new_changelog'
 def set_cl(inst, basedn, log, args):
     cl = Changelog5(inst)
     attrs = _args_to_attrs(args)
@@ -504,6 +519,7 @@ def set_cl(inst, basedn, log, args):
     log.info("Successfully updated replication changelog")
 
 
+# This subcommand is available when 'not ds_supports_new_changelog'
 def get_cl(inst, basedn, log, args):
     cl = Changelog5(inst)
     if args and args.json:
@@ -511,6 +527,49 @@ def get_cl(inst, basedn, log, args):
     else:
         log.info(cl.display())
 
+# This subcommand is available when 'ds_supports_new_changelog'
+# that means there is a changelog config entry per backend (aka suffix)
+def set_per_backend_cl(inst, basedn, log, args):
+    suffix = args.suffix
+    cl = Changelog(inst, suffix)
+    attrs = _args_to_attrs(args)
+    replace_list = []
+    did_something = False
+
+    if args.encrypt:
+        cl.replace('nsslapd-encryptionalgorithm', 'AES')
+        del args.encrypt
+        did_something = True
+        log.info("You must restart the server for this to take effect")
+    elif args.disable_encrypt:
+        cl.remove_all('nsslapd-encryptionalgorithm')
+        del args.disable_encrypt
+        did_something = True
+        log.info("You must restart the server for this to take effect")
+
+    for attr, value in attrs.items():
+        if value == "":
+            cl.remove_all(attr)
+            did_something = True
+        else:
+            replace_list.append((attr, value))
+    if len(replace_list) > 0:
+        cl.replace_many(*replace_list)
+    elif not did_something:
+        raise ValueError("There are no changes to set for the replication changelog")
+
+    log.info("Successfully updated replication changelog")
+
+# This subcommand is available when 'ds_supports_new_changelog'
+# that means there is a changelog config entry per backend (aka suffix)
+def get_per_backend_cl(inst, basedn, log, args):
+    suffix = args.suffix
+    cl = Changelog(inst, suffix)
+    if args and args.json:
+        log.info(cl.get_all_attrs_json())
+    else:
+        log.info(cl.display())
+
 
 def create_repl_manager(inst, basedn, log, args):
     manager_name = "replication manager"
@@ -688,8 +747,26 @@ def add_agmt(inst, basedn, log, args):
     if args.strip_list is not None:
         properties['nsds5replicastripattrs'] = args.strip_list
 
-    # We do need the bind dn and credentials for none-sasl bind methods
-    if (bind_method in ('simple', 'sslclientauth')) and (args.bind_dn is None or args.bind_passwd is None):
+    # Handle the optional bootstrap settings
+    if args.bootstrap_bind_dn is not None:
+        if not is_a_dn(args.bootstrap_bind_dn):
+            raise ValueError("The replica bootstrap bind DN is not a valid DN")
+        properties['nsDS5ReplicaBootstrapBindDN'] = args.bootstrap_bind_dn
+    if args.bootstrap_bind_passwd is not None:
+        properties['nsDS5ReplicaBootstrapCredentials'] = args.bootstrap_bind_passwd
+    if args.bootstrap_bind_method is not None:
+        bs_bind_method = args.bootstrap_bind_method.lower()
+        if bs_bind_method != "simple" and bs_bind_method != "sslclientauth":
+            raise ValueError('Bootstrap bind method can only be "SIMPLE" or "SSLCLIENTAUTH"')
+        properties['nsDS5ReplicaBootstrapBindMethod'] = args.bootstrap_bind_method
+    if args.bootstrap_conn_protocol is not None:
+        bootstrap_conn_protocol = args.bootstrap_conn_protocol.lower()
+        if bootstrap_conn_protocol != "ldap" and bootstrap_conn_protocol != "ldaps" and bootstrap_conn_protocol != "starttls":
+            raise ValueError('Bootstrap connection protocol can only be "LDAP", "LDAPS", or "STARTTLS"')
+        properties['nsDS5ReplicaBootstrapTransportInfo'] = args.bootstrap_conn_protocol
+
+    # We do need the bind dn and credentials for 'simple' bind method
+    if (bind_method == 'simple') and (args.bind_dn is None or args.bind_passwd is None):
         raise ValueError("You need to set the bind dn (--bind-dn) and the password (--bind-passwd) for bind method ({})".format(bind_method))
 
     # Create the agmt
@@ -759,6 +836,14 @@ def set_agmt(inst, basedn, log, args):
                 for frac_attr in value.split():
                     frac_list += " " + frac_attr
                 value = frac_list
+            elif attr == 'nsds5replicabootstrapbindmethod':
+                bs_bind_method = value.lower()
+                if bs_bind_method != "simple" and bs_bind_method != "sslclientauth":
+                    raise ValueError('Bootstrap bind method can only be "SIMPLE" or "SSLCLIENTAUTH"')
+            elif attr == 'nsds5replicabootstraptransportinfo':
+                bs_conn_protocol = value.lower()
+                if bs_conn_protocol != "ldap" and bs_conn_protocol != "ldaps" and bs_conn_protocol != "starttls":
+                    raise ValueError('Bootstrap bind method can only be "LDAP", "LDAPS, or "STARTTLS"')
             modlist.append((attr, value))
 
     if len(modlist) > 0:
@@ -1038,61 +1123,96 @@ def list_abort_cleanallruv(inst, basedn, log, args):
             log.info("No CleanAllRUV abort tasks found")
 
 
-def dump_cl(inst, basedn, log, args):
-    if args.output_file:
-        fh = logging.FileHandler(args.output_file, mode='w')
-        log.addHandler(fh)
+def dump_def_cl(inst, basedn, log, args):
     replicas = Replicas(inst)
+    try:
+        replica = replicas.get(args.replica_root)
+    except:
+        raise ValueError(f"Suffix \"{args.replica_root}\" is not enabled for replication")
+    try:
+        replica_name = replica.get_attr_val_utf8_l("nsDS5ReplicaName")
+        ldif_dir = inst.get_ldif_dir()
+        replica.begin_task_cl2ldif()
+        if not replica.task_finished():
+            raise ValueError("The changelog export task (CL2LDIF) did not complete in time")
+        log.info(f'Successfully created: ' + os.path.join(ldif_dir, f'{replica_name}_cl.ldif'))
+    except:
+        raise ValueError("Failed to export replication changelog")
+
+def dump_cl(inst, basedn, log, args):
     if not args.changelog_ldif:
-        replicas.process_and_dump_changelog(replica_roots=args.replica_roots,
+        replicas = Replicas(inst)
+        replicas.process_and_dump_changelog(replica_root=args.replica_root,
+                                            output_file=args.output_file,
                                             csn_only=args.csn_only,
                                             preserve_ldif_done=args.preserve_ldif_done,
-                                            log=log)
+                                            decode=args.decode)
     else:
+        # Modify an existing LDIF file
         try:
             assert os.path.exists(args.changelog_ldif)
         except AssertionError:
             raise FileNotFoundError(f"File {args.changelog_ldif} was not found")
-        cl_ldif = ChangelogLDIF(args.changelog_ldif, log)
+        cl_ldif = ChangelogLDIF(args.changelog_ldif, output_file=args.output_file)
         if args.csn_only:
             cl_ldif.grep_csn()
         else:
             cl_ldif.decode()
 
+def restore_cl_def_ldif(inst, basedn, log, args):
+    """
+    Import the server default cl ldif files from the server's LDIF directory
+    """
+    ldif_dir = inst.get_ldif_dir()
+    replicas = Replicas(inst)
+    try:
+        replica = replicas.get(args.replica_root)
+    except:
+        raise ValueError(f"Suffix \"{args.replica_root}\" is not enabled for replication")
+    replica_name = replica.get_attr_val_utf8_l("nsDS5ReplicaName")
+    target_ldif = f'{replica_name}_cl.ldif'
+    target_ldif_exists = os.path.exists(os.path.join(ldif_dir, target_ldif))
+    if not target_ldif_exists:
+        # We are trying to import the default ldif, but it's not there
+        raise ValueError(f'The default LDAP file "{target_ldif}" does not exist')
+    # Import the default LDIF
+    replicas.restore_changelog(replica_root=args.replica_root, log=log)
 
 def restore_cl_ldif(inst, basedn, log, args):
-    user_ldif = os.path.abspath(args.LDIF_PATH[0])
-    try:
-        assert os.path.exists(user_ldif)
-    except AssertionError:
-        raise FileNotFoundError(f"File {args.LDIF_PATH[0]} was not found")
+    user_ldif = None
+    if args.LDIF_PATH[0]:
+        user_ldif = os.path.abspath(args.LDIF_PATH[0])
+        try:
+            assert os.path.exists(user_ldif)
+        except AssertionError:
+            raise FileNotFoundError(f"File {args.LDIF_PATH[0]} was not found")
 
+    ldif_dir = inst.get_ldif_dir()
     replicas = Replicas(inst)
-    replica = replicas.get(args.replica_root[0])
+    try:
+        replica = replicas.get(args.replica_root)
+    except:
+        raise ValueError(f"Suffix \"{args.replica_root}\" is not enabled for replication")
     replica_name = replica.get_attr_val_utf8_l("nsDS5ReplicaName")
-    target_ldif = f'{replica_name}.ldif'
+    target_ldif = os.path.join(ldif_dir, f'{replica_name}_cl.ldif')
     target_ldif_exists = os.path.exists(target_ldif)
-    cl = Changelog5(inst)
-    cl_dir = cl.get_attr_val_utf8_l("nsslapd-changelogdir")
-    cl_dir_file = [i.lower() for i in os.listdir(cl_dir) if i.lower().startswith(replica_name)][0]
-    cl_dir_stat = os.stat(os.path.join(cl_dir, cl_dir_file))
+
     # Make sure we don't remove existing files
     if target_ldif_exists:
         copy_with_permissions(target_ldif, f'{target_ldif}.backup')
     copyfile(user_ldif, target_ldif)
-    os.chown(target_ldif, cl_dir_stat[stat.ST_UID], cl_dir_stat[stat.ST_GID])
-    os.chmod(target_ldif, cl_dir_stat[stat.ST_MODE])
-    replicas.restore_changelog(replica_roots=args.replica_root, log=log)
+
+    ldif_dir_file = [i.lower() for i in os.listdir(ldif_dir) if i.lower().startswith(replica_name)][0]
+    ldif_dir_stat = os.stat(os.path.join(ldif_dir, ldif_dir_file))
+    os.chown(target_ldif, ldif_dir_stat[stat.ST_UID], ldif_dir_stat[stat.ST_GID])
+    os.chmod(target_ldif, ldif_dir_stat[stat.ST_MODE])
+    replicas.restore_changelog(replica_root=args.replica_root, log=log)
     os.remove(target_ldif)
     if target_ldif_exists:
+        # restore the original file that we backed up
         os.rename(f'{target_ldif}.backup', target_ldif)
 
 
-def restore_cl_dir(inst, basedn, log, args):
-    replicas = Replicas(inst)
-    replicas.restore_changelog(replica_roots=args.REPLICA_ROOTS, log=log)
-
-
 def create_parser(subparsers):
 
     ############################################
@@ -1105,8 +1225,8 @@ def create_parser(subparsers):
     repl_enable_parser = repl_subcommands.add_parser('enable', help='Enable replication for a suffix')
     repl_enable_parser.set_defaults(func=enable_replication)
     repl_enable_parser.add_argument('--suffix', required=True, help='The DN of the suffix to be enabled for replication')
-    repl_enable_parser.add_argument('--role', required=True, help="The Replication role: \"master\", \"hub\", or \"consumer\"")
-    repl_enable_parser.add_argument('--replica-id', help="The replication identifier for a \"master\".  Values range from 1 - 65534")
+    repl_enable_parser.add_argument('--role', required=True, help="The Replication role: \"supplier\", \"hub\", or \"consumer\"")
+    repl_enable_parser.add_argument('--replica-id', help="The replication identifier for a \"supplier\".  Values range from 1 - 65534")
     repl_enable_parser.add_argument('--bind-group-dn', help="A group entry DN containing members that are \"bind/supplier\" DNs")
     repl_enable_parser.add_argument('--bind-dn', help="The Bind or Supplier DN that can make replication updates")
     repl_enable_parser.add_argument('--bind-passwd', help="Password for replication manager(--bind-dn).  This will create the manager entry if a value is set")
@@ -1134,11 +1254,11 @@ def create_parser(subparsers):
     repl_winsync_status_parser.add_argument('--bind-dn', help="The DN to use to authenticate to the consumer")
     repl_winsync_status_parser.add_argument('--bind-passwd', help="The password for the bind DN")
 
-    repl_promote_parser = repl_subcommands.add_parser('promote', help='Promte replica to a Hub or Master')
+    repl_promote_parser = repl_subcommands.add_parser('promote', help='Promote replica to a Hub or Supplier')
     repl_promote_parser.set_defaults(func=promote_replica)
     repl_promote_parser.add_argument('--suffix', required=True, help="The DN of the replication suffix to promote")
-    repl_promote_parser.add_argument('--newrole', required=True, help='Promote this replica to a \"hub\" or \"master\"')
-    repl_promote_parser.add_argument('--replica-id', help="The replication identifier for a \"master\".  Values range from 1 - 65534")
+    repl_promote_parser.add_argument('--newrole', required=True, help='Promote this replica to a \"hub\" or \"supplier\"')
+    repl_promote_parser.add_argument('--replica-id', help="The replication identifier for a \"supplier\".  Values range from 1 - 65534")
     repl_promote_parser.add_argument('--bind-group-dn', help="A group entry DN containing members that are \"bind/supplier\" DNs")
     repl_promote_parser.add_argument('--bind-dn', help="The Bind or Supplier DN that can make replication updates")
 
@@ -1159,59 +1279,65 @@ def create_parser(subparsers):
 
     repl_demote_parser = repl_subcommands.add_parser('demote', help='Demote replica to a Hub or Consumer')
     repl_demote_parser.set_defaults(func=demote_replica)
-    repl_demote_parser.add_argument('--suffix', required=True, help="Promte this replica to a \"hub\" or \"consumer\"")
+    repl_demote_parser.add_argument('--suffix', required=True, help="Promote this replica to a \"hub\" or \"consumer\"")
     repl_demote_parser.add_argument('--newrole', required=True, help="The Replication role: \"hub\", or \"consumer\"")
 
     repl_get_parser = repl_subcommands.add_parser('get', help='Get replication configuration')
     repl_get_parser.set_defaults(func=get_repl_config)
     repl_get_parser.add_argument('--suffix', required=True, help='Get the replication configuration for this suffix DN')
 
-    repl_create_cl = repl_subcommands.add_parser('create-changelog', help='Create the replication changelog')
-    repl_create_cl.set_defaults(func=create_cl)
+    repl_set_per_backend_cl = repl_subcommands.add_parser('set-changelog', help='Set replication changelog attributes.')
+    repl_set_per_backend_cl.set_defaults(func=set_per_backend_cl)
+    repl_set_per_backend_cl.add_argument('--suffix', required=True, help='The suffix that uses the changelog')
+    repl_set_per_backend_cl.add_argument('--max-entries', help="The maximum number of entries to get in the replication changelog")
+    repl_set_per_backend_cl.add_argument('--max-age', help="The maximum age of a replication changelog entry")
+    repl_set_per_backend_cl.add_argument('--trim-interval', help="The interval to check if the replication changelog can be trimmed")
+    repl_set_per_backend_cl.add_argument('--encrypt', action='store_true', help="Set the replication changelog to use encryption.  You must export & import the changelog after setting this.")
+    repl_set_per_backend_cl.add_argument('--disable-encrypt', action='store_true', help="Set the replication changelog to not use encryption.  You must export & import the changelog after setting this.")
 
-    repl_delete_cl = repl_subcommands.add_parser('delete-changelog', help='Delete the replication changelog.  This will invalidate any existing replication agreements')
-    repl_delete_cl.set_defaults(func=delete_cl)
+    repl_get_per_backend_cl = repl_subcommands.add_parser('get-changelog', help='Display replication changelog attributes.')
+    repl_get_per_backend_cl.set_defaults(func=get_per_backend_cl)
+    repl_get_per_backend_cl.add_argument('--suffix', required=True, help='The suffix that uses the changelog')
 
-    repl_set_cl = repl_subcommands.add_parser('set-changelog', help='Set replication changelog attributes.')
-    repl_set_cl.set_defaults(func=set_cl)
-    repl_set_cl.add_argument('--cl-dir', help="The replication changelog location on the filesystem")
-    repl_set_cl.add_argument('--max-entries', help="The maximum number of entries to get in the replication changelog")
-    repl_set_cl.add_argument('--max-age', help="The maximum age of a replication changelog entry")
-    repl_set_cl.add_argument('--compact-interval', help="The replication changelog compaction interval")
-    repl_set_cl.add_argument('--trim-interval', help="The interval to check if the replication changelog can be trimmed")
+    repl_export_cl = repl_subcommands.add_parser('export-changelog', help='Export the Directory Server replication changelog to an LDIF')
+    export_subcommands = repl_export_cl.add_subparsers(help='Export Replication Changelog')
+    repl_export_cl = export_subcommands.add_parser('to-ldif', help='Export the specific single LDIF file.  '
+                                                                   'This is typically used for setting up changelog encryption')
+    repl_export_cl.set_defaults(func=dump_cl)
+    repl_export_cl.add_argument('-c', '--csn-only', action='store_true',
+                                help="Export and interpret CSN only. This option can be used with or without -i option.  "
+                                     "The LDIF file that is generated can not be imported and is only used debugging purposes")
+    repl_export_cl.add_argument('-d', '--decode', action='store_true',
+                                help="Decode the base64 values in each changelog entry.  "
+                                     "The LDIF file that is generated can not be imported and is only used debugging purposes")
+    repl_export_cl.add_argument('-l', '--preserve-ldif-done', action='store_true',
+                              help="Preserve generated ldif.done files in changelog dirextory.")
+    repl_export_cl.add_argument('-i', '--changelog-ldif',
+                                help="If you already have a changelog LDIF file, but the changes in that file are encoded,"
+                                     " you may use this option to decode the changes in that LDIF file.")
+    repl_export_cl.add_argument('-o', '--output-file', required=True, help="Path name for the final result.")
+    repl_export_cl.add_argument('-r', '--replica-root', required=True,
+                                help="Specify replica root whose changelog you want to export.")
 
-    repl_get_cl = repl_subcommands.add_parser('get-changelog', help='Display replication changelog attributes.')
-    repl_get_cl.set_defaults(func=get_cl)
+    repl_def_export_cl = export_subcommands.add_parser('default', help='Export the replication changelog to the server\'s default LDIF directory.')
+    repl_def_export_cl.set_defaults(func=dump_def_cl)
+    repl_def_export_cl.add_argument('-r', '--replica-root', required=True,
+                                    help="Specify replica root whose changelog you want to export.")
 
-    repl_dump_cl = repl_subcommands.add_parser('dump-changelog', help='Decode Directory Server replication change log and dump it to an LDIF')
-    repl_dump_cl.set_defaults(func=dump_cl)
-    repl_dump_cl.add_argument('-c', '--csn-only', action='store_true',
-                              help="Dump and interpret CSN only. This option can be used with or without -i option.")
-    repl_dump_cl.add_argument('-l', '--preserve-ldif-done', action='store_true',
-                              help="Preserve generated ldif.done files from changelogdir.")
-    repl_dump_cl.add_argument('-i', '--changelog-ldif',
-                              help="If you already have a ldif-like changelog, but the changes in that file are encoded,"
-                                   " you may use this option to decode that ldif-like changelog. It should be base64 encoded.")
-    repl_dump_cl.add_argument('-o', '--output-file', help="Path name for the final result. Default to STDOUT if omitted.")
-    repl_dump_cl.add_argument('-r', '--replica-roots', nargs="+",
-                              help="Specify replica roots whose changelog you want to dump. The replica "
-                                   "roots may be seperated by comma. All the replica roots would be dumped if the option is omitted.")
+    repl_import_cl = repl_subcommands.add_parser('import-changelog',
+        help='Restore/Import Directory Server replication change log from an LDIF file.  This is typically used when managing changelog encryption')
+    import_subcommands = repl_import_cl.add_subparsers(help='Restore/Import Replication Changelog')
+    import_ldif = import_subcommands.add_parser('from-ldif', help='Restore/Import a specific single LDIF file.')
+    import_ldif.set_defaults(func=restore_cl_ldif)
+    import_ldif.add_argument('LDIF_PATH', nargs=1, help='The path of the changelog LDIF file.')
+    import_ldif.add_argument('-r', '--replica-root', required=True,
+                             help="Specify the replica root whose changelog you want to import.")
 
-    repl_restore_cl = repl_subcommands.add_parser('restore-changelog',
-                                                  help='Restore Directory Server replication change log from LDIF file or change log directory')
-    restore_subcommands = repl_restore_cl.add_subparsers(help='Restore Replication Changelog')
-    restore_ldif = restore_subcommands.add_parser('from-ldif', help='Restore a single LDIF file.')
-    restore_ldif.set_defaults(func=restore_cl_ldif)
-    restore_ldif.add_argument('LDIF_PATH', nargs=1, help='The path of changelog LDIF file.')
-    restore_ldif.add_argument('-r', '--replica-root', nargs=1, required=True,
-                              help="Specify one replica root whose changelog you want to restore. "
-                                   "The replica root will be consumed from the LDIF file name if the option is omitted.")
-
-    restore_changelogdir = restore_subcommands.add_parser('from-changelogdir', help='Restore LDIF files from changelogdir.')
-    restore_changelogdir.set_defaults(func=restore_cl_dir)
-    restore_changelogdir.add_argument('REPLICA_ROOTS', nargs="+",
-                                      help="Specify replica roots whose changelog you want to restore. The replica "
-                                           "roots may be seperated by comma. All the replica roots would be dumped if the option is omitted.")
+    import_def_ldif = import_subcommands.add_parser('default',
+        help='Import the default changelog LDIF file created by the server.')
+    import_def_ldif.set_defaults(func=restore_cl_def_ldif)
+    import_def_ldif.add_argument('-r', '--replica-root', required=True,
+                                 help="Specify the replica root whose changelog you want to import.")
 
     repl_set_parser = repl_subcommands.add_parser('set', help='Set an attribute in the replication configuration')
     repl_set_parser.set_defaults(func=set_repl_config)
@@ -1231,7 +1357,7 @@ def create_parser(subparsers):
                                                             "while waiting to acquire the consumer.  Default is 300 seconds")
     repl_set_parser.add_argument('--repl-backoff-min', help="The starting time in seconds a replication agreement should stay in a backoff state "
                                                             "while waiting to acquire the consumer.  Default is 3 seconds")
-    repl_set_parser.add_argument('--repl-release-timeout', help="A timeout in seconds a replication master should send "
+    repl_set_parser.add_argument('--repl-release-timeout', help="A timeout in seconds a replication supplier should send "
                                                                 "updates before it yields its replication session")
 
     repl_monitor_parser = repl_subcommands.add_parser('monitor', help='Get the full replication topology report')
@@ -1318,7 +1444,7 @@ def create_parser(subparsers):
     agmt_add_parser.add_argument('--strip-list', help="A list of attributes that are removed from updates only if the event "
                                                       "would otherwise be empty.  Typically this is set to \"modifiersname\" and \"modifytimestmap\"")
     agmt_add_parser.add_argument('--schedule', help="Sets the replication update schedule: 'HHMM-HHMM DDDDDDD'  D = 0-6 (Sunday - Saturday).")
-    agmt_add_parser.add_argument('--conn-timeout', help="The timeout used for replicaton connections")
+    agmt_add_parser.add_argument('--conn-timeout', help="The timeout used for replication connections")
     agmt_add_parser.add_argument('--protocol-timeout', help="A timeout in seconds on how long to wait before stopping "
                                                             "replication when the server is under load")
     agmt_add_parser.add_argument('--wait-async-results', help="The amount of time in milliseconds the server waits if "
@@ -1329,6 +1455,10 @@ def create_parser(subparsers):
     agmt_add_parser.add_argument('--session-pause-time', help="The amount of time in seconds a supplier should wait between update sessions.")
     agmt_add_parser.add_argument('--flow-control-window', help="Sets the maximum number of entries and updates sent by a supplier, which are not acknowledged by the consumer.")
     agmt_add_parser.add_argument('--flow-control-pause', help="The time in milliseconds to pause after reaching the number of entries and updates set in \"--flow-control-window\"")
+    agmt_add_parser.add_argument('--bootstrap-bind-dn', help="An optional Bind DN the agreement can use to bootstrap initialization when bind groups are being used")
+    agmt_add_parser.add_argument('--bootstrap-bind-passwd', help="The bootstrap credentials for the Bind DN")
+    agmt_add_parser.add_argument('--bootstrap-conn-protocol', help="The replication bootstrap connection protocol: LDAP, LDAPS, or StartTLS")
+    agmt_add_parser.add_argument('--bootstrap-bind-method', help="The bind method: \"SIMPLE\", or \"SSLCLIENTAUTH\"")
     agmt_add_parser.add_argument('--init', action='store_true', default=False, help="Initialize the agreement after creating it.")
 
     # Set - Note can not use add's parent args because for "set" there are no "required=True" args
@@ -1347,7 +1477,7 @@ def create_parser(subparsers):
     agmt_set_parser.add_argument('--strip-list', help="A list of attributes that are removed from updates only if the event "
                                                       "would otherwise be empty.  Typically this is set to \"modifiersname\" and \"modifytimestmap\"")
     agmt_set_parser.add_argument('--schedule', help="Sets the replication update schedule: 'HHMM-HHMM DDDDDDD'  D = 0-6 (Sunday - Saturday).")
-    agmt_set_parser.add_argument('--conn-timeout', help="The timeout used for replicaton connections")
+    agmt_set_parser.add_argument('--conn-timeout', help="The timeout used for replication connections")
     agmt_set_parser.add_argument('--protocol-timeout', help="A timeout in seconds on how long to wait before stopping "
                                                             "replication when the server is under load")
     agmt_set_parser.add_argument('--wait-async-results', help="The amount of time in milliseconds the server waits if "
@@ -1357,6 +1487,10 @@ def create_parser(subparsers):
     agmt_set_parser.add_argument('--session-pause-time', help="The amount of time in seconds a supplier should wait between update sessions.")
     agmt_set_parser.add_argument('--flow-control-window', help="Sets the maximum number of entries and updates sent by a supplier, which are not acknowledged by the consumer.")
     agmt_set_parser.add_argument('--flow-control-pause', help="The time in milliseconds to pause after reaching the number of entries and updates set in \"--flow-control-window\"")
+    agmt_set_parser.add_argument('--bootstrap-bind-dn', help="An optional Bind DN the agreement can use to bootstrap initialization when bind groups are being used")
+    agmt_set_parser.add_argument('--bootstrap-bind-passwd', help="The bootstrap credentials for the Bind DN")
+    agmt_set_parser.add_argument('--bootstrap-conn-protocol', help="The replication bootstrap connection protocol: LDAP, LDAPS, or StartTLS")
+    agmt_set_parser.add_argument('--bootstrap-bind-method', help="The bind method: \"SIMPLE\", or \"SSLCLIENTAUTH\"")
 
     # Get
     agmt_get_parser = agmt_subcommands.add_parser('get', help='Get replication configuration')
